#include "DAQChannelMap/DAQChannelMap.h"
#include "DAQChannelMap/BitFields.h"
#include "DAQChannelMap/boundscheck.h"

#include <stdio.h>
#include <iostream>

using namespace daqchannelmap;

/**********************************************************************************
 * DAQChannelMap Class Implimentation                                             *
 **********************************************************************************/

// Initialize the singleton pointer to NULL
DAQChannelMap*  DAQChannelMap::_instanceID = NULL;

//------------------------------------------------------------------------------
/*! getInstance method returns the current instance if there is one
  or creates the initial instance if there is not one */
DAQChannelMap* DAQChannelMap::getInstance(){
  if(_instanceID == NULL){	               // Check if there is an instance active
    _instanceID = new DAQChannelMap();         // If there is not, make a new one
  }
  return _instanceID;
}

//------------------------------------------------------------------------------
DAQChannelMap::DAQChannelMap(){ }// end of DAQChannelMap constructor
DAQChannelMap::~DAQChannelMap(){ }		//!< Default Destructor

//------------------------------------------------------------------------------
dchan DAQChannelMap::encodeDChan(DetectorIdent detector, diblock_t diblock,
				 dcm_id_t dcmid,    dcm_link_t dcmport,
                                 pixel_t pixel) const{
  dchan chan = 0x0;
  BITFIELDSET(chan,detector, MASK_DETECTOR, SHIFT_DETECTOR);
  BITFIELDSET(chan,diblock,  MASK_DIBLOCK,  SHIFT_DIBLOCK);
  BITFIELDSET(chan,dcmid,    MASK_DCMID,    SHIFT_DCMID);
  BITFIELDSET(chan,dcmport,  MASK_DCMPORT,  SHIFT_DCMPORT);
  BITFIELDSET(chan,pixel,    MASK_PIXELID,  SHIFT_PIXELID);
  return chan;
}

//------------------------------------------------------------------------------
lchan DAQChannelMap::encodeLChan(DetectorIdent   detector,
				 offline_plane_t plane_0,
				 offline_cell_t  cell_0) const{
  lchan chan = 0x0;
  BITFIELDSET(chan,detector, MASK_OFFLINE_DETECTOR, SHIFT_OFFLINE_DETECTOR);
  BITFIELDSET(chan,plane_0 , MASK_OFFLINE_PLANE   , SHIFT_OFFLINE_PLANE);
  BITFIELDSET(chan,cell_0  , MASK_OFFLINE_CELL    , SHIFT_OFFLINE_CELL);
  return chan;
}

//------------------------------------------------------------------------------
dchan DAQChannelMap::encodeDChan(lchan chan) {

  const DetectorIdent det       = getDetectorOffline(chan);
  const diblock_t     diblock   = getDiBlockOffline(chan);
  const dcm_id_t      dcm_id    = getDCMOffline(chan);
  const dcm_link_t    dcm_link  = getDCMLinkOffline(chan);
  const pixel_t       pixel     = getPixelOffline(chan);

  const apd_pixel_t   apd_pixel = pixel2APDPixel(pixel);
  
  return encodeDChan(det, diblock, dcm_id, dcm_link, apd_pixel);
}

//------------------------------------------------------------------------------
lchan DAQChannelMap::encodeLChan(dchan chan) {
  const DetectorIdent   det     = getDetector(chan);
  const offline_plane_t plane_0 = getPlane(chan);
  const offline_cell_t  cell_0  = getCell(chan);

  return encodeLChan(det, plane_0, cell_0);
}

//------------------------------------------------------------------------------
DetectorIdent DAQChannelMap::getDetector(dchan daqchan) const{
  return (DetectorIdent)BOUNDSCHECK(
		     BITFIELDGET(daqchan,MASK_DETECTOR,SHIFT_DETECTOR),
		     MIN_DET_ID, MAX_DET_ID, ERROR_UNKNOWN_DET);
}

//------------------------------------------------------------------------------
diblock_t DAQChannelMap::getDiBlock(dchan daqchan) const{
  return BOUNDSCHECK(
		     BITFIELDGET(daqchan,MASK_DIBLOCK,SHIFT_DIBLOCK),
		     MIN_DIBLOCK, MAX_DIBLOCK, ERROR_DIBLOCK_OUTRANGE);
}

//------------------------------------------------------------------------------
dcm_id_t DAQChannelMap::getDCM(dchan daqchan) const{
  return  BOUNDSCHECK(
		     BITFIELDGET(daqchan,MASK_DCMID,SHIFT_DCMID),
		     MIN_DCM,MAX_DCM, ERROR_DCM_OUTRANGE);
}

//------------------------------------------------------------------------------
dcm_link_t DAQChannelMap::getDCMLink(dchan daqchan) const {
  return BOUNDSCHECK(
		     BITFIELDGET(daqchan,MASK_DCMPORT,SHIFT_DCMPORT),
		     MIN_PORT,MAX_PORT, ERROR_PORT_OUTRANGE);
}

//------------------------------------------------------------------------------
pixel_t DAQChannelMap::getPixel(dchan daqchan) const {
  return BOUNDSCHECK(
		     BITFIELDGET(daqchan,MASK_PIXELID,SHIFT_PIXELID),
		     MIN_PIXEL,MAX_PIXEL, ERROR_PIXEL_OUTRANGE);
}

//------------------------------------------------------------------------------
DetView_TYPE DAQChannelMap::getView(dchan daqchan) const {
  const DetectorIdent detector_id  = getDetector(daqchan);
  const dcm_id_t      dcm_id       = getDCM(daqchan);
  const DiBlock_TYPE  diblock_type = getDiBlockType(daqchan);

  const DetPlane_TYPE plane_type   = getPlaneType(detector_id, dcm_id, diblock_type);

  return detPlaneToDetView(plane_type);
}

//------------------------------------------------------------------------------
//! Returns the type of diblock that the channel was on
DiBlock_TYPE DAQChannelMap::getDiBlockType(dchan daqchan) const {
    const DetectorIdent detector       = getDetector(daqchan);
    const diblock_t     diblock_number = getDiBlock(daqchan);

    return getDiBlockTypeBase(detector, diblock_number);
}

//------------------------------------------------------------------------------
//! Returns the Plane in coordinates relative to the dcm (i.e. 0-31)
offline_plane_t DAQChannelMap::getPlane(dchan daqchan) {

    const DetectorIdent detector_id = getDetector(daqchan);
    const diblock_t     diblock_id  = getDiBlock(daqchan);
    const dcm_id_t      dcm_id      = getDCM(daqchan);
    const dcm_link_t    dcm_link_id = getDCMLink(daqchan);

    /// Now Based on (detector_id, diblock_id, dcm_id, dcm_link_id), calculate these variables
    const DiBlock_TYPE    diblock_type                = getDiBlockTypeBase(detector_id, diblock_id);
    const offline_plane_t local_plane_in_diblock_view = getLocalPlaneInDiblockView(dcm_link_id, diblock_type);
    const DetPlane_TYPE   plane_type                  = getPlaneType(detector_id, dcm_id, diblock_type);

    /// Local plane in diblock
    offline_plane_t offline_plane_id = getLocalPlaneInDiblock(local_plane_in_diblock_view, diblock_type, plane_type);

    /// Count all planes before this Diblock. Add them to the offline plane number
    for(diblock_t idiblock = 1; idiblock < diblock_id; idiblock++){
        offline_plane_id += getNumberOfPlanesInDiblock(detector_id, idiblock);
    }// end of loop over diblocks before the current one

    return offline_plane_id;
}

//------------------------------------------------------------------------------
// Offline Channel Decoding
DetectorIdent DAQChannelMap::getDetectorOffline(lchan chan) const {
  return (DetectorIdent)BOUNDSCHECK(
		     BITFIELDGET(chan,MASK_OFFLINE_DETECTOR,SHIFT_OFFLINE_DETECTOR),
		     MIN_DET_ID, MAX_DET_ID, ERROR_UNKNOWN_DET);
}

//------------------------------------------------------------------------------
offline_plane_t DAQChannelMap::getPlaneOffline(lchan chan) const {
  return BOUNDSCHECK(
		     BITFIELDGET(chan, MASK_OFFLINE_PLANE, SHIFT_OFFLINE_PLANE),
		     0,NUM_PLANES_DIBLOCK[getDetectorOffline(chan)]*NUM_DIBLOCK_DET[getDetectorOffline(chan)], ERROR_PLANE_OUTRANGE);
}

//------------------------------------------------------------------------------
offline_cell_t DAQChannelMap::getCellOffline(lchan chan) const {
  return BOUNDSCHECK(
		     BITFIELDGET(chan,MASK_OFFLINE_CELL,SHIFT_OFFLINE_CELL),
	 	     0,NUM_CELLS*NUM_PLANES_DIBLOCK[getDetectorOffline(chan)]*NUM_DIBLOCK_DET[getDetectorOffline(chan)], ERROR_CELL_OUTRANGE);
}

//------------------------------------------------------------------------------
uint32_t DAQChannelMap::getBlockOffline(lchan chan) const {

  const DetectorIdent   detector = getDetectorOffline(chan);
  const offline_plane_t plane_0  = getPlaneOffline(chan); // plane starts from 0 in Offline

  return getBlockNumber(detector, plane_0);
}
//------------------------------------------------------------------------------
uint32_t DAQChannelMap::getDiBlockOffline(lchan chan) const {
  const DetectorIdent   detector = getDetectorOffline(chan);
  const offline_plane_t plane_0  = getPlaneOffline(chan); // plane starts from 0 in Offline

  return getDiblockNumber(detector, plane_0);
}

//------------------------------------------------------------------------------
//! Returns the module relative to the current dcm that a channel belongs to (0 or 1)
//  Module 0 refers to DCM ports in range 0-15, 32-47 (uppermost)
//  Module 1 refers to DCM ports in range 16-31, 48-63 
//------------------------------------------------------------------------------
uint32_t DAQChannelMap::getLocalModuleInDCM(dcm_link_t dcm_link, DiBlock_TYPE diblock_type) const {

    /// Muon catcher screws things up
    if(diblock_type == MC_DIBLOCK)
        return dcm_link / 16;

    return (dcm_link % 32) / 16;
}

//------------------------------------------------------------------------------
//! Returns the Cell in coordinates relative to the detector view (i.e. 1-"a big number" for far det)
//  Cells are referenced in the offline 0-95 (NDOS) with cell 0 at the snout end in X-View
//  And cell 0 at the bottom of the detector
uint32_t DAQChannelMap::getCell(dchan daqchan) const {

    const dcm_id_t      dcm_id       = getDCM(daqchan);
    const dcm_link_t    dcm_link_id  = getDCMLink(daqchan);
    const apd_pixel_t   apd_pixel_id = getPixel(daqchan);
    const DetectorIdent detector_id  = getDetector(daqchan);
    const DiBlock_TYPE  diblock_type = getDiBlockType(daqchan);

    const pixel_t       pixel_id     = APDPixel2Pixel(apd_pixel_id);

    const online_cell_t online_cell  = getOnlineCell(detector_id, dcm_id, dcm_link_id, pixel_id, diblock_type);
    const DetPlane_TYPE plane_type   = getPlaneType(detector_id, dcm_id, diblock_type);

    return onlineCellToOfflineCell(online_cell, detector_id, plane_type);
}

//------------------------------------------------------------------------------
DiBlock_TYPE DAQChannelMap::getDiBlockTypeOffline(lchan chan) const {
  const diblock_t     diblock_id = getDiBlockOffline(chan);
  const DetectorIdent detector   = getDetectorOffline(chan);

  return getDiBlockTypeBase(detector, diblock_id);
}

//------------------------------------------------------------------------------
DetBlock_TYPE DAQChannelMap::getBlockType(lchan logicalchan) const {
    const DetectorIdent detector      = getDetectorOffline(logicalchan);
    const uint32_t      block_number  = getBlockOffline(logicalchan);

    return getBlockType(block_number, detector);
}

//------------------------------------------------------------------------------
DetBlock_TYPE DAQChannelMap::getBlockType(block_t blockID, DetectorIdent detector) const  {

    const diblock_t diblockID                     = 1 + ((blockID-1)/2); //< we count diblocks from 1
    const uint32_t  local_block_number_in_diblock = EVEN(blockID);       //< we count from 0

    const DiBlock_TYPE  diblock_type = getDiBlockTypeBase(detector, diblockID);
    const DetBlock_TYPE block_type   = getBlockTypeBase(diblock_type, local_block_number_in_diblock);

    return block_type;
}

//------------------------------------------------------------------------------
// Returns local plane number for the diblock from offline channel
offline_plane_t DAQChannelMap::getLocalPlaneInDiblockOffline(lchan chan) const {
  const DetectorIdent   detector = getDetectorOffline(chan);
  const offline_plane_t plane_0  = getPlaneOffline(chan);    //< plane starts from 0 in Offline

  return getLocalPlaneInDiblockBase(detector, plane_0);
}

//------------------------------------------------------------------------------
offline_plane_t DAQChannelMap::getLocalPlaneInBlockOffline(lchan chan) const {
  const DetectorIdent   detector = getDetectorOffline(chan);
  const offline_plane_t plane_0  = getPlaneOffline(chan); // plane starts from 0 in Offline

  return getLocalPlaneInBlockBase(detector, plane_0);
}

//------------------------------------------------------------------------------
DetView_TYPE DAQChannelMap::getViewOffline(lchan chan) const {
    return detPlaneToDetView(getPlaneTypeOffline(chan));
}

//------------------------------------------------------------------------------
DetPlane_TYPE DAQChannelMap::getPlaneTypeOffline(lchan chan) const {
  const DetBlock_TYPE block_type           = getBlockType(chan);
  const uint32_t      local_plane_in_block = getLocalPlaneInBlockOffline(chan);

  const DetPlane_TYPE start_plane_type     = getStartPlaneTypeInBlock(block_type);

  /// If local plane is even, then the view is the same as the one that the block starts with
  if(EVEN(local_plane_in_block)) return start_plane_type;

  return getOtherPlaneType(start_plane_type);
}

//------------------------------------------------------------------------------
offline_plane_t DAQChannelMap::getLocalPlaneInDiblockViewOffline(lchan chan) {

    const offline_plane_t local_plane_in_block = getLocalPlaneInBlockOffline(chan);

    offline_plane_t local_plane_in_block_view  = local_plane_in_block / 2;
    if(getLocalBlockInDiblockOffline(chan) == 0) return local_plane_in_block_view;

    /// otherwise, we need to calculate how many planes of the same view
    /// were in the first diblock

    const DetectorIdent detector = getDetectorOffline(chan);

    /// The first block of the diblock - it's the one before the current
    const block_t       block_id_of_the_first_block_in_diblock   = getBlockOffline(chan) - 1;
    const DetBlock_TYPE block_type_of_the_first_block_in_diblock = getBlockType(block_id_of_the_first_block_in_diblock, detector);

    /// View of the plane
    const DetPlane_TYPE plane_type = getPlaneTypeOffline(chan);

    /// Add the number of planes in the view for that diblock
    local_plane_in_block_view += getNumberOfPlaneTypeInBlock(block_type_of_the_first_block_in_diblock, plane_type);

    return local_plane_in_block_view;
}

//------------------------------------------------------------------------------
// Returns the plane number of the channel in the X or Y view
offline_plane_t DAQChannelMap::getPlaneViewOffline(lchan chan){
  
  const diblock_t     diblock    = getDiBlockOffline(chan);
  const DetectorIdent detector   = getDetectorOffline(chan);
  const DetPlane_TYPE plane_type = getPlaneTypeOffline(chan);

  // Sum the number of planes in this view prior to the current diblock
  // that the channel is from

  /// the plane number that the current channel is on in the local diblock
  uint32_t total_planes = getLocalPlaneInDiblockViewOffline(chan);

  for(uint32_t idiblock = 0; idiblock < diblock; ++idiblock){
    const DiBlock_TYPE diblock_type = getDiBlockTypeBase(detector, idiblock);
    total_planes +=  getNumberOfPlaneTypeInDiblock(diblock_type, plane_type);
  }// end of loop over diblocks

  return total_planes;
}

//------------------------------------------------------------------------------
dcm_id_t DAQChannelMap::getDCMOffline(lchan chan) const {

  const online_cell_t online_cell_0     = getOnlineCellOffline(chan);
  const dcm_id_t      first_dcm_in_view = getFirstDCMInViewOffline(chan);
  const DiBlock_TYPE  diblock_type      = getDiBlockTypeOffline(chan);

  return first_dcm_in_view + online_cell_0 / getNumberOfCellsPerDCM(diblock_type);
}// end of getDCMOffline

//------------------------------------------------------------------------------
dcm_link_t DAQChannelMap::getDCMLinkOffline(lchan chan) {

    const uint32_t        local_module                = getLocalModuleInDCMOffline(chan);
    const offline_plane_t local_plane_in_diblock_view = getLocalPlaneInDiblockViewOffline(chan);

    /// Calculate DCM link number according to NOvA-doc-4219
    dcm_link_t dcm_link = local_plane_in_diblock_view
                        + 16*local_module;
    if(local_plane_in_diblock_view > 15) dcm_link += 16;

    return dcm_link;
}// end of getDCMLinkOffline

//------------------------------------------------------------------------------
// If it's the upstream block in diblock, then it's 0, if not, it's 1
uint32_t DAQChannelMap::getLocalBlockInDiblockOffline(lchan logicalchan) const {
    const uint32_t block = getBlockOffline(logicalchan);

    return EVEN(block);
}// end of getLocalBlockInDiblockOffline

//------------------------------------------------------------------------------
dchan DAQChannelMap::l2d(offline_plane_t plane, offline_cell_t cell, DetectorIdent detector)  {return encodeDChan(encodeLChan(detector,plane,cell));};

//------------------------------------------------------------------------------
/// First DCM in the view. if the channel is in a YZ plane, the DCM is 1. Otherwise, it's 7 for the far detector and 2 for NDOS
dcm_id_t DAQChannelMap::getFirstDCMInViewOffline(lchan chan) const {

    const DetPlane_TYPE plane_type   = getPlaneTypeOffline(chan);
    const DetectorIdent detector     = getDetectorOffline(chan);
    const DiBlock_TYPE  diblock_type = getDiBlockTypeOffline(chan);

    return getFirstDCMInView(detector, plane_type, diblock_type);
}

//------------------------------------------------------------------------------
/// First DCM in the view. if the channel is in a YZ plane, the DCM is 1. Otherwise, it's 7 for the far detector and 2 for NDOS
dcm_id_t DAQChannelMap::getFirstDCMInView(DetectorIdent detector, DetPlane_TYPE plane_type, DiBlock_TYPE diblock_type) const {

    /// If it's XZ plane, then the first DCM is 1
    if(plane_type == XZ_PLANE) return 1;

    /// Otherwise, it's the next DCM after the last XZ DCM
    return getNumberOfXDCM(detector, diblock_type) + 1;
}

//------------------------------------------------------------------------------
dcm_id_t  DAQChannelMap::getLastDCMInViewOffline(lchan chan) const {

    const DetPlane_TYPE plane_type   = getPlaneTypeOffline(chan);
    const DetectorIdent detector     = getDetectorOffline(chan);
    const DiBlock_TYPE  diblock_type = getDiBlockTypeOffline(chan);

    return getLastDCMInView(detector, plane_type, diblock_type);
}
//------------------------------------------------------------------------------
dcm_id_t  DAQChannelMap::getLastDCMInView(DetectorIdent detector, DetPlane_TYPE plane_type, DiBlock_TYPE diblock_type) const {


    /// If it's XZ plane, then the first DCM is NUM_X_DCMS
    if(plane_type == XZ_PLANE) return getNumberOfXDCM(detector, diblock_type);

    /// So, it's vertical (YZ) plane

    return getFirstDCMInView(detector, plane_type, diblock_type) + getNumberOfYDCM(detector, diblock_type) - 1;
}

//------------------------------------------------------------------------------
pixel_t  DAQChannelMap::getPixelOffline(lchan chan) const {
    const online_cell_t online_cell_0     = getOnlineCellOffline(chan);
    const DiBlock_TYPE  diblock_type      = getDiBlockTypeOffline(chan);

    const online_cell_t local_cell_in_dcm = getLocalCellInDCM(online_cell_0, diblock_type);

    return local_cell_in_dcm % NUM_PIXELS;
}

//------------------------------------------------------------------------------
uint32_t DAQChannelMap::getLocalModuleInDCMOffline(lchan chan) const {

    const online_cell_t online_cell_0  = getOnlineCellOffline(chan);
    const DiBlock_TYPE  diblock_type   = getDiBlockTypeOffline(chan);

    return getLocalModuleInDCMOffline(online_cell_0, diblock_type);
}

//------------------------------------------------------------------------------
uint32_t DAQChannelMap::getLocalModuleInDCMOffline(online_cell_t online_cell_0, DiBlock_TYPE diblock_type) const{

    /// This is what it should be. But Muon catcher screws things up
    online_cell_t local_cell_in_dcm = getLocalCellInDCM(online_cell_0, diblock_type);

    return local_cell_in_dcm / NUM_PIXELS;
}

//------------------------------------------------------------------------------
online_cell_t DAQChannelMap::getLocalCellInDCM(online_cell_t online_cell_0, DiBlock_TYPE diblock_type)const{

    const online_cell_t number_of_cells_per_dcm = getNumberOfCellsPerDCM(diblock_type);

    return online_cell_0 - (online_cell_0 / number_of_cells_per_dcm)*number_of_cells_per_dcm;
}

//------------------------------------------------------------------------------
DetPlane_TYPE DAQChannelMap::getPlaneType(DetectorIdent detector_id, dcm_id_t dcm_id, DiBlock_TYPE diblock_type) const {
        if(dcm_id <= getNumberOfXDCM(detector_id, diblock_type)) return XZ_PLANE;

        return YZ_PLANE;
}

//------------------------------------------------------------------------------
offline_plane_t DAQChannelMap::getLocalPlaneInDiblockView(dcm_link_t dcm_link_id, DiBlock_TYPE diblock_type) const {

    /// DCM links are counted from 0. Refer to NOvA-doc-4219.
    if(dcm_link_id < 32) return dcm_link_id % 16;

    /// Muon catcher screws things up. We use 1 DCM for muon catcher in each view
    if(diblock_type == MC_DIBLOCK) return (dcm_link_id - 32) % 16;

    return 16 + (dcm_link_id - 32) % 16;

}// end of getLocalPlaneInDiblockView

//------------------------------------------------------------------------------
uint32_t DAQChannelMap::getLocalBlockInDiblock(DiBlock_TYPE diblock_type,
                                               offline_plane_t local_plane_in_diblock_in_view,
                                               DetPlane_TYPE plane_type){

    const DetBlock_TYPE first_block_of_diblock_type = getBlockTypeBase(diblock_type, 0);

    /// Get Number of planes of the same type in that first diblock.
    /// Needed to figure out which block current plane belongs to
    const uint32_t number_of_planes_of_the_same_type_in_first_block_of_the_diblock = getNumberOfPlaneTypeInBlock(first_block_of_diblock_type, plane_type);

    if(local_plane_in_diblock_in_view < number_of_planes_of_the_same_type_in_first_block_of_the_diblock) return 0;

    return 1;
}

//------------------------------------------------------------------------------
offline_plane_t DAQChannelMap::getLocalPlaneInDiblock(offline_plane_t local_plane_in_diblock_view,
                                                      DiBlock_TYPE diblock_type,
                                                      DetPlane_TYPE plane_type){

    const uint32_t local_block_in_diblock  = getLocalBlockInDiblock(diblock_type, local_plane_in_diblock_view, plane_type);

    const DetBlock_TYPE current_block_type = getBlockTypeBase(diblock_type, local_block_in_diblock);
    const DetPlane_TYPE start_plane_type   = getStartPlaneTypeInBlock(current_block_type);

    /// Does the block starts with a different plane?
    uint32_t is_block_starts_with_different_plane = 0;
    if(plane_type != start_plane_type) is_block_starts_with_different_plane = 1;

    /// If it's the first block. We can calculate plabe easily
    if(local_block_in_diblock == 0) return local_plane_in_diblock_view*2 + is_block_starts_with_different_plane;

    /// So, it's the second block. Need to do some magic

    /// First block type
    const DetBlock_TYPE   first_block_type          = getBlockTypeBase(diblock_type, 0);

    /// To find local plane in the (second) block in the view we subtract
    /// number of planes of the same type in the first diblock from the local plane in diblock view
    const offline_plane_t local_plane_in_block_view = local_plane_in_diblock_view - getNumberOfPlaneTypeInBlock(first_block_type, plane_type);

    return getNumberOfPlanesInBlock(first_block_type) + local_plane_in_block_view*2 + is_block_starts_with_different_plane;
}

//------------------------------------------------------------------------------
online_cell_t DAQChannelMap::getOnlineCell(DetectorIdent detector, 
                                           dcm_id_t      dcm_id,
                                           dcm_link_t    dcm_link,
                                           pixel_t       pixel,
                                           DiBlock_TYPE  diblock_type) const {

    const DetPlane_TYPE plane_type          = getPlaneType(detector, dcm_id, diblock_type);
    const dcm_id_t      first_dcm_in_view   = getFirstDCMInView(detector, plane_type, diblock_type);
    const uint32_t      local_module_in_dcm = getLocalModuleInDCM(dcm_link, diblock_type);

    return getNumberOfCellsPerDCM(diblock_type) * (dcm_id - first_dcm_in_view)
           + NUM_PIXELS * local_module_in_dcm
           + pixel;
}

//------------------------------------------------------------------------------
online_cell_t DAQChannelMap::getOnlineCellOffline(lchan chan) const {

    const offline_cell_t cell_offline_0 = getCellOffline(chan);
    const DetPlane_TYPE  plane_type     = getPlaneTypeOffline(chan);
    const DetectorIdent  detector       = getDetectorOffline(chan);

    return offlineCellToOnlineCell(cell_offline_0, detector, plane_type);
}

//------------------------------------------------------------------------------
offline_cell_t DAQChannelMap::onlineCellToOfflineCell(online_cell_t online_cell_0,
                                                      DetectorIdent detector,
                                                      DetPlane_TYPE plane_type
                                                      ) const {
    return ConvertCell((uint32_t)online_cell_0, detector, plane_type);
}
//------------------------------------------------------------------------------
online_cell_t DAQChannelMap::offlineCellToOnlineCell(offline_cell_t offline_cell_0,
                                                     DetectorIdent detector,
                                                     DetPlane_TYPE plane_type
                                                     ) const {
    return ConvertCell((uint32_t)offline_cell_0, detector, plane_type);
}

//------------------------------------------------------------------------------
uint32_t DAQChannelMap::ConvertCell(uint32_t input_cell, 
                                    DetectorIdent detector,
                                    DetPlane_TYPE plane_type) const {

    /// If the plane is vertical, then the order is reversed
    if(plane_type == YZ_PLANE){
        return getNumberOfChannelsInYPlane(detector) - input_cell - 1;
    }// end of if the plane is horizontal

    /// If the plane is horizontal, then the order is the same
    return input_cell;
}

//------------------------------------------------------------------------------
bool DAQChannelMap::print(dchan chan){

  const DetectorIdent   det      = getDetector(chan);
  const diblock_t       diblock  = getDiBlock(chan);
  const dcm_id_t        dcm_id   = getDCM(chan);
  const dcm_link_t      dcm_link = getDCMLink(chan);
  const pixel_t         pixel    = getPixel(chan);

  const offline_plane_t plane_0  = getPlane(chan);
  const offline_cell_t  cell_0   = getCell(chan);

  const lchan offline_channel = encodeLChan(det, plane_0, cell_0);

  std::cout<<"DAQChannel=";printHex(chan);
  std::cout<<"(";printBinary(chan);std::cout<<")";
  std::cout<<"\n";

  std::cout<<" Detector="<<det<<"("<<getDetectorLabel(det)<<")";
  std::cout<<" Diblock="<<diblock;
  std::cout<<" DCM="<<dcm_id;
  std::cout<<" DCMLink="<<dcm_link;
  std::cout<<" Pixel="<<pixel;

  std::cout<<"               ";

  std::cout<<" Plane="<<plane_0;
  std::cout<<" Cell="<<cell_0;

  std::cout<<"\n";
  std::cout<<" OfflineChannel=";
  printHex(offline_channel);std::cout<<"(";printBinary(offline_channel);std::cout<<")";

  std::cout<<"\n";

  return true;
}

//------------------------------------------------------------------------------
bool DAQChannelMap::printOffline(lchan chan){

  const DetectorIdent   det      = getDetectorOffline(chan);
  const offline_plane_t plane_0  = getPlaneOffline(chan);
  const offline_cell_t  cell_0   = getCellOffline(chan);

  const diblock_t     diblock  = getDiBlockOffline(chan);
  const dcm_id_t      dcm_id   = getDCMOffline(chan);
  const dcm_link_t    dcm_link = getDCMLinkOffline(chan);
  const pixel_t       pixel    = getPixelOffline(chan);
  
  const dchan daq_channel   = encodeDChan(det, diblock, dcm_id, dcm_link, pixel);

  std::cout<<"OfflineChannel=";printHex(chan);
  std::cout<<"(";printBinary(chan);std::cout<<")";
  std::cout<<"\n";

  std::cout<<"Detector="<<det<<"("<<getDetectorLabel(det)<<")";
  std::cout<<" Plane="<<plane_0;
  std::cout<<" Cell="<<cell_0;

  std::cout<<"               ";

  std::cout<<" Diblock="<<diblock;
  std::cout<<" DCM="<<dcm_id;
  std::cout<<" DCMLink="<<dcm_link;
  std::cout<<" Pixel="<<pixel;

  std::cout<<"\n";
  std::cout<<"DAQChannel=";printHex(daq_channel);
  std::cout<<"(";printBinary(daq_channel);std::cout<<")";

  std::cout<<"\n";

  return true;
}

//------------------------------------------------------------------------------
// Stolen from RawDAQData::printWord
bool  DAQChannelMap::printBinary(uint32_t value){

        const int SHIFT = 8 * sizeof( unsigned ) - 1;
        const unsigned MASK = 1 << SHIFT;

        for ( int i = 1; i <= SHIFT + 1; i++ ){
                std::cout << (  value & MASK ? '1' : '0' );
                value <<= 1;

                if ( i % 8 == 0 ) std::cout << ' ';
        }
        
        return true;
}// end of printBinary

//------------------------------------------------------------------------------
// Stolen from RawDAQData::printWord
bool DAQChannelMap::printHex(uint32_t value){
    printf("0x%08x",value);

    return true;
}// end of printHex


